#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 13 21:23:38 2019

@author: cdholmes
"""
import numpy as np

def funcerror_uncorr(func,x0,xerr,h=None,**kwargs):
    '''Propagate uncertainties in function parameters to uncertainty in function value. 
    
    The uncertainty will be evaluated at f(x0). The method uses 
    a numerical estimate of the derivative of f(x0). For non-linear functions
    the method is appropriate for small relative errors (i.e. small xerr/x0).
    The method neglects possible error correlations between multiple parameters 
    in x0.
    The function variables must all be continuous variables 
    (no categorical or discrete variables).
    
    Example. To find the uncertainty in f(x,y) at for values x=1, y=2, where 
    the uncertainty (e.g. standard error) in x is 0.1 and the uncertainty in y 
    is 0.3...
    $ print( funcerr_uncorr( f, [1,2], [0.1,0.3] ) )
    
    Parameters
    ----------
    func : function handle
        function to be evaluated
    x0   : list or tuple
        parameter values for func
    xerr : list or tuple
        uncertainty in parameter values x0
    h    : float (default=1e-3)
        fractional perturbation to x0 used to estimate slope of func via finite difference 
    
    Returns
    -------
    ferr : float
        the uncertainty in f(x0) given xerr uncertainty in x0
    '''

    # Increment size for finite difference
    # Numerical Methods in C recommends h * x, where x is the variable and 
    # h = sqrt(machine precision). For a 64-bit (double) float, that implies 
    # h = 2e-8. For a 32-bit (single) float, that implies h = 3e-4.
    # We choose a larger value to reduce roundoff error. Note 
    # that very large h causes formula error where secant line is not parallel 
    # to instantaneous slope.
    if h is None:
        h = 1e-7
        h = 1e-3

    ferr = np.float64(0)

    for i,xi in enumerate(x0):
        xplus     = x0.copy()
        xminus    = x0.copy()
        xplus[i]  = xplus[i]  + h*xi
        xminus[i] = xminus[i] - h*xi

        # Centered finite difference estimate of derivative
        dfdx = ( func(*xplus,**kwargs) - func(*xminus,**kwargs) ) / ( xplus[i]-xminus[i] )

        ferr += dfdx**2 * xerr[i]**2

    ferr = np.sqrt(ferr)

    return ferr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''Read and write ICARTT (ffi1001) format files

Created on Mon Aug 13 16:59:25 2018
@author: C.D. Holmes
'''

import os
import pandas as pd


def read_icartt( files, usePickle=False, timeIndex=False ):
    '''Read ICARTT file or files into a pandas DataFrame
    
    Parameters
    ----------
    files : list or str
        path to ICARTT file or files that will be read. 
        Data within these files will be concatenated, so files should all contain the same variables
    usePickle : bool (default=False)
        if usePickle=True, the data will be written to a pkl file with ".pkl" appended to path
        On subsequent read_icartt calls, data will be read from the .pkl file, if it exists
    timeIndex : bool (default=False)
        sets DataFrame index to the time variable from the ICARTT file, rather than a row counter
        
    Returns
    -------
    obs : pandas.DataFrame
        ICARTT file contents. In addition to column names for the ICARTT variables, 
        the DataFrame columns also include 'time' in pandas.DatetimeIndex format and 
        'file' that is the ordinal number of the input file that each row was read from
    '''

    # Files input must be string or list of strings
    if isinstance( files, str ):
        # Convert to list
        files = [files]
    elif isinstance( files, list ):
        # Do nothing
        pass
    else:
        raise TypeError( 'read_icartt: files must be a filename or list of filenames' )

    obsall = []
    for n,file in enumerate(files):

        # Read from Pickle file, if it exists and is requested
        pklfile = file+'.pkl'
        if (usePickle and os.path.isfile(pklfile)):
            obs = pd.read_pickle(pklfile)

        else:

            # Ensure file exists
            if not os.path.isfile( file ):
                raise FileNotFoundError( file+" doesn't exist" )

            # Read the ICARTT file
            with open( file, 'r', encoding='ascii' ) as f:

                # Read the number of header lines
                nheader, fmt = f.readline().split(',')
                nheader = int(nheader)

                # Ensure this is a 
                if int(fmt) != 1001:
                    raise Exception( 'read_icartt: '+file+' is not an ICARTT (ffi1001) file' )

                # Skip 5 lines
                for junk in range(5):
                    next(f)

                # Read date
                year, month, day = map( int, f.readline().split(',')[0:3] )

                # Skip line
                next(f)

                # Read name of the time variable
                tname, tunit = [s.strip() for s in f.readline().split(',')[0:2]]

                # Raise exception if the time unit is not seconds;
                # may need to be handled differently below
                if (tunit in ['s','seconds','seconds (from midnight UTC)','seconds_past_midnight']):
                    # Use unit expected by pandas
                    tunit = 's'
                else:
                    print(tunit)
                    raise Exception( 'read_icartt: time expected in seconds (s); '+
                                     'unit in file: ',tunit )

                # Number of dependent variables
                nvar = int( f.readline() )

                # Scale factor for dependent variables
                scale = [ float(s) for s in f.readline().split(',') ]

                # Missing value flag for dependent variables
                naflag = [ s.strip() for s in f.readline().split(',') ]

                # Dependent variable names from the next nvar lines
                varnames = [ f.readline().split(',')[0] for v in range(nvar) ]

                # Missing flags for each variable as a dict
                nadict = { varnames[i]: naflag[i] for i in range(nvar) }

            # Read data
            obs = pd.read_csv(file, skiprows=nheader-1,
                              na_values=nadict, skipinitialspace=True)

            # Catch missing data that are not reported as integers
            #obs[obs==-99999] = np.nan

            # Strip whitespace from column names
            #obs.columns = obs.columns.str.strip()

            # Apply scale factors
            for i, s in enumerate(scale):
                if s != 1:
                    obs[varnames[i]] *= s

            # Add a time variable in datetime format
            obs['time'] = pd.DatetimeIndex( pd.Timestamp(year=year,month=month,day=day) +
                                         pd.TimedeltaIndex(obs[tname],tunit) )

            # Add flight number
            obs['file'] = n+1

            # Use time variable for index
            if timeIndex:
                obs.index = obs.time

            # Save pickle
            if usePickle:
                obs.to_pickle(pklfile)

        # Add to list
        obsall.append(obs)

    # Concatenate all files into one dataframe
    obs = pd.concat( obsall, sort=True )

    return obs

def write_icartt(filename, df, **kwargs):
    '''Write an ICARTT ffi1001 file

    Arguments
    ---------
    filename : str
        File to be created
    df : Pandas dataframe
        df should contain data and metadata
        df attributes must include all of the "required" and "normal_comments" names.
        The INDEPENDENT_VARIABLE_DEFINITION and DEPENDENT_VARIABLE_DEFINITION should be dicts containing
        {'VariableName':'units, standard name, description'}
        Only variables listed in INDEPENDENT_VARIABLE_DEFINITION and DEPENDENT_VARIABLE_DEFINITION
        will be written to the output file.
    **kwargs
        passed to pandas.to_csv
    '''

    normal_comments = ['PI_CONTACT_INFO',
                'PLATFORM',
                'LOCATION',
                'ASSOCIATED_DATA',
                'INSTRUMENT_INFO',
                'DATA_INFO',
                'UNCERTAINTY',
                'ULOD_FLAG',
                'ULOD_VALUE',
                'LLOD_FLAG',
                'LLOD_VALUE',
                'DM_CONTACT_INFO',
                'PROJECT_INFO',
                'STIPULATIONS_ON_USE',
                'OTHER_COMMENTS',
                'REVISION',
                'REVISION_COMMENTS']

    required = ['PI_NAME',
                'ORGANIZATION_NAME',
                'SOURCE_DESCRIPTION',
                'MISSION_NAME',
                'VOLUME_INFO',
                'DATE_LINE',
                'TIME_INTERVAL',
                'INDEPENDENT_VARIABLE_DEFINITION',
                'NUMBER_DEPENDENT_VARIABLES',
                'DEPENDENT_SCALE_LINE',
                'DEPENDENT_MISSING_FLAGS',
                'DEPENDENT_VARIABLE_DEFINITION',
                'SPECIAL_COMMENTS',
                'NORMAL_COMMENTS']

    # Variables that will be written to file
    ictvars = list(df.INDEPENDENT_VARIABLE_DEFINITION.keys()) + \
              list(df.DEPENDENT_VARIABLE_DEFINITION.keys())

    # Form the header
    header = []

    for k in required:
        # Special handling for some 
        if k=='VOLUME_INFO':
            header.append( '1, 1')
        elif k=='DATE_LINE':
            header.append( df.time.iloc[0].strftime('%Y, %m, %d, ') + 
                           pd.Timestamp.today().strftime('%Y, %m, %d'))
        elif k=='TIME_INTERVAL':
            header.append( '1'  ) # ***UPDATE LATER: TIME INTERVAL IN SECONDS
        elif k=='INDEPENDENT_VARIABLE_DEFINITION':
            keydict = getattr( df, k )
            header.append( list(keydict.keys())[0] + ', ' + list(keydict.values())[0] )            
        elif k=='DEPENDENT_VARIABLE_DEFINITION':
            keydict = getattr( df, k )
            nvars = len(keydict)
            header.append( str(nvars) )
            header.append( ','.join(['1']*nvars) ) # Scale line
            header.append( ','.join(['-9999']*nvars) ) # Missing flags
            for kn in keydict.keys():
                header.append( kn + ', ' + keydict[kn])
        elif k in ['TIME_INTERVAL',
                'NUMBER_DEPENDENT_VARIABLES', 
                'DEPENDENT_SCALE_LINE',
                'DEPENDENT_MISSING_FLAGS']:
            #*** Handled elsewhere
            pass
        elif k=='SPECIAL_COMMENTS':
            v = getattr( df, k )
            if v:
                # Expect a string or array of several lines
                header.append( str(len(list(v))) )
                header.extend( list(v) )
            else:
                header.append( '0' )
        elif k=='NORMAL_COMMENTS':
            # Form the comment block
            nc= []
            for kn in normal_comments:
                v = getattr( df, kn )
                if kn=='REVISION_COMMENTS':
                    # Expect a string or array of several lines
                    nc.extend( list(v) )
                else:    
                    nc.append( '{:s}: {:s}'.format(kn,v) )
            nc.append( ', '.join(ictvars)) # Also add list of variable names

            # Add normal comments to the header
            header.append( str(len(nc)) )
            header.extend( nc )
        else:
            v = getattr( df, k )
            header.append( v )

    # Write the file
    with open(filename,'w',encoding='ascii') as f:
        f.write('{:d}, 1001\n'.format(len(header)+1))  # +1 accounts for this line
        for line in header:
            f.write(line+'\n')
        df[ictvars].to_csv( f,
                            index=False,
                            header=False,
                            na_rep='-9999',
                            **kwargs )



"""Data analysis programs from the ACGC research group"""

def package_version(package_name):
    '''Find version string for package name'''
    from importlib.metadata import version, PackageNotFoundError
    try:
        result = version(package_name)
    except PackageNotFoundError:
        result = "unknown version"
    return result

__version__ = package_version('acgc')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov  7 17:01:36 2018

@author: cdholmes
"""

import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import numpy as np

# Artificial data with n latitudes, m longitudes, starting at latitude lat1 and ending at lat2 
def create_data(n,m,lat1,lat2):
    # edges of the latitude-longitude grid
    latedge = np.linspace(lat1,lat2,n+1)
    lonedge = np.linspace(-180,180,m+1)
    
    # Centers of the latitude-longitude grid
    lat      = np.convolve(latedge, [0.5,0.5], 'valid' )
    lon      = np.convolve(lonedge, [0.5,0.5], 'valid' )
    lon2d, lat2d = np.meshgrid( lon, lat )

    # Artificial data
    data = np.exp( -lat2d**2/90**2 -lon2d**2/120**2 )

    return latedge, lonedge, data    


##########################################################
# Set up map and plot data

# Open a figure
plt.figure(1)
plt.clf()

# number of points
n = 20
m = 40

# data coordinates: Use PlateCarree for data in lat-lon coordinates
datacoord = ccrs.PlateCarree()

##########################################################
# Latitude range 90S-90N; This FAILS

# Data from 90S to 90N
latedge, lonedge, data = create_data(n,m,-90,90)

# Map projection 
ax = plt.subplot(1,3,1, projection=ccrs.LambertConformal())
ax.coastlines()

# Add data
plt.pcolormesh(lonedge,latedge,data,transform=datacoord)

plt.title('LambertConformal FAILS')

##########################################################
# Same data, different projection


# Map projection 
ax = plt.subplot(1,3,2, projection=ccrs.AlbersEqualArea())
ax.coastlines()

# Add data
plt.pcolormesh(lonedge,latedge,data,transform=datacoord)

plt.title('Albers WORKS (same data)')

##########################################################
# Latitude range 89.999S-90N; This WORKS

# Data from 90S to 90N
latedge, lonedge, data = create_data(n,m,-89.999,90)

# Map projection 
ax = plt.subplot(1,3,3, projection=ccrs.LambertConformal())
ax.coastlines()

plt.pcolormesh(lonedge,latedge,data,transform=datacoord)

plt.title('LambertConformal WORKS (limit 89.999S)')

plt.show()

#plt.savefig('test.png')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''A collection of statistical methods'''

from .bivariate import *
from .bivariate_lines import *
from .boxcar import *
from .partial_corr import *
from .weighted import *

# Other sub-modules that are not imported because of
# better options or limited applications
# from .tapply import tapply    # use groupby in pandas and xarray
# from .sma_warton_fit import * # Use sma
try:
    # Requires R and rpy2
    from .loess import loess
except ModuleNotFoundError:
    # Skip
    pass




# -*- coding: utf-8 -*-
"""
Created on Mon Jun 15 17:43:20 2015

@author: cdholmes
"""

#conda install cartopy

import cartopy.crs as ccrs
import cartopy.feature as cfeature 

import matplotlib.pyplot as plt
import numpy as np

# Create artificial point data
# number of points
npoint = 50
# random latitude, longitude and value
plat  = np.random.rand(npoint) * 180 - 90
plon  = np.random.rand(npoint) * 360 - 180
pdata = np.random.rand(npoint) 

# Clear the figure
plt.figure(1)
plt.clf()

# data coordinates: Use PlateCarree for data in lat-lon coordinates
datacoord = ccrs.PlateCarree()

# Projection
projcoord = ccrs.PlateCarree()
projcoord = ccrs.Robinson()

#Setup map projection and continent outlines
ax = plt.axes(projection=projcoord)

# Change map extent, if needed
#ax.set_extent([-180, 180, -90, 90], ccrs.PlateCarree())

# Add data points
plt.scatter( plon, plat, c=pdata, s=50, marker='o', transform=datacoord )

# draw parallels and meridians, but don't bother labelling them.
ax.gridlines(draw_labels=False)

# Add coasts and states; Use ‘110m’ or ‘50m’ resolution
ax.coastlines(resolution='110m')

# Fill the continents
ax.add_feature(cfeature.LAND, facecolor='0.75')

# add colorbar
plt.colorbar()#location='bottom')
plt.clim(0,1)

# Optional: Convert the points into the projection coordinate system
xyz = projcoord.transform_points( datacoord, plon, plat )
# Optional: Plot the points using projection coordinates; Note: Now we don't need "transform"
x=xyz[:,0]; y=xyz[:,1]
plt.scatter( x, y, s=50, marker='o', facecolor='none', edgecolor='black' )

plt.title('Artificial data')
#plt.show()



#!/usr/bin/env python3

def test_acgc():
    import acgc 

def test_erroranalysis():
    import acgc.erroranalysis

def test_figstyle():
    import acgc.figstyle

def test_gctools():
    import acgc.gctools

def test_hytools():
    import acgc.hytools

def test_icartt():
    import acgc.icartt

def test_igra():
    import acgc.igra

def test_map_scalebar():
    import acgc.map_scalebar

def test_mettools():
    import acgc.mettools

def test_modetools():
    import acgc.modetools

def test_nctools():
    import acgc.nctools

def test_solar():
    import acgc.solar

def test_stats():
    import acgc.stats

def test_timetools():
    import acgc.timetools



#!/usr/bin/env python3
""" Scale bar from StackOverflow
https://stackoverflow.com/questions/32333870/how-can-i-show-a-km-ruler-on-a-cartopy-matplotlib-plot
"""

import numpy as np
import cartopy.crs as ccrs
import cartopy.geodesic as cgeo


def _axes_to_lonlat(ax, coords):
    """(lon, lat) from axes coordinates."""
    display = ax.transAxes.transform(coords)
    data = ax.transData.inverted().transform(display)
    lonlat = ccrs.PlateCarree().transform_point(*data, ax.projection)

    return lonlat


def _upper_bound(start, direction, distance, dist_func):
    """A point farther than distance from start, in the given direction.

    It doesn't matter which coordinate system start is given in, as long
    as dist_func takes points in that coordinate system.

    Args:
        start:     Starting point for the line.
        direction  Nonzero (2, 1)-shaped array, a direction vector.
        distance:  Positive distance to go past.
        dist_func: A two-argument function which returns distance.

    Returns:
        Coordinates of a point (a (2, 1)-shaped NumPy array).
    """
    if distance <= 0:
        raise ValueError(f"Minimum distance is not positive: {distance}")

    if np.linalg.norm(direction) == 0:
        raise ValueError("Direction vector must not be zero.")

    # Exponential search until the distance between start and end is
    # greater than the given limit.
    length = 0.1
    end = start + length * direction

    while dist_func(start, end) < distance:
        length *= 2
        end = start + length * direction

    return end


def _distance_along_line(start, end, distance, dist_func, tol):
    """Point at a distance from start on the segment  from start to end.

    It doesn't matter which coordinate system start is given in, as long
    as dist_func takes points in that coordinate system.

    Args:
        start:     Starting point for the line.
        end:       Outer bound on point's location.
        distance:  Positive distance to travel.
        dist_func: Two-argument function which returns distance.
        tol:       Relative error in distance to allow.

    Returns:
        Coordinates of a point (a (2, 1)-shaped NumPy array).
    """
    initial_distance = dist_func(start, end)
    if initial_distance < distance:
        raise ValueError(f"End is closer to start ({initial_distance}) than "
                         f"given distance ({distance}).")

    if tol <= 0:
        raise ValueError(f"Tolerance is not positive: {tol}")

    # Binary search for a point at the given distance.
    left = start
    right = end

    while not np.isclose(dist_func(start, right), distance, rtol=tol):
        midpoint = (left + right) / 2

        # If midpoint is too close, search in second half.
        if dist_func(start, midpoint) < distance:
            left = midpoint
        # Otherwise the midpoint is too far, so search in first half.
        else:
            right = midpoint

    return right


def _point_along_line(ax, start, distance, angle=0, tol=0.01):
    """Point at a given distance from start at a given angle.

    Args:
        ax:       CartoPy axes.
        start:    Starting point for the line in axes coordinates.
        distance: Positive physical distance to travel.
        angle:    Anti-clockwise angle for the bar, in radians. Default: 0
        tol:      Relative error in distance to allow. Default: 0.01

    Returns:
        Coordinates of a point (a (2, 1)-shaped NumPy array).
    """
    # Direction vector of the line in axes coordinates.
    direction = np.array([np.cos(angle), np.sin(angle)])

    geodesic = cgeo.Geodesic()

    # Physical distance between points.
    def dist_func(a_axes, b_axes):
        a_phys = _axes_to_lonlat(ax, a_axes)
        b_phys = _axes_to_lonlat(ax, b_axes)

        # Geodesic().inverse returns a NumPy MemoryView like [[distance,
        # start azimuth, end azimuth]].
        return geodesic.inverse(a_phys, b_phys).base[0, 0]

    end = _upper_bound(start, direction, distance, dist_func)

    return _distance_along_line(start, end, distance, dist_func, tol)


def scale_bar(ax, location, length, metres_per_unit=1000, unit_name='km',
              tol=0.01, angle=0, color='black', linewidth=3, text_offset=0.005,
              ha='center', va='bottom', plot_kwargs=None, text_kwargs=None,
              **kwargs):
    """Add a scale bar to CartoPy axes.

    For angles between 0 and 90 the text and line may be plotted at
    slightly different angles for unknown reasons. To work around this,
    override the 'rotation' keyword argument with text_kwargs.

    Args:
        ax:              CartoPy axes.
        location:        Position of left-side of bar in axes coordinates.
        length:          Geodesic length of the scale bar.
        metres_per_unit: Number of metres in the given unit. Default: 1000
        unit_name:       Name of the given unit. Default: 'km'
        tol:             Allowed relative error in length of bar. Default: 0.01
        angle:           Anti-clockwise rotation of the bar.
        color:           Color of the bar and text. Default: 'black'
        linewidth:       Same argument as for plot.
        text_offset:     Perpendicular offset for text in axes coordinates.
                         Default: 0.005
        ha:              Horizontal alignment. Default: 'center'
        va:              Vertical alignment. Default: 'bottom'
        **plot_kwargs:   Keyword arguments for plot, overridden by **kwargs.
        **text_kwargs:   Keyword arguments for text, overridden by **kwargs.
        **kwargs:        Keyword arguments for both plot and text.
    """
    # Setup kwargs, update plot_kwargs and text_kwargs.
    if plot_kwargs is None:
        plot_kwargs = {}
    if text_kwargs is None:
        text_kwargs = {}

    plot_kwargs = {'linewidth': linewidth, 'color': color, **plot_kwargs,
                   **kwargs}
    text_kwargs = {'ha': ha, 'va': va, 'rotation': angle, 'color': color,
                   **text_kwargs, **kwargs}

    # Convert all units and types.
    location = np.asarray(location)  # For vector addition.
    length_metres = length * metres_per_unit
    angle_rad = angle * np.pi / 180

    # End-point of bar.
    end = _point_along_line(ax, location, length_metres, angle=angle_rad,
                            tol=tol)

    # Coordinates are currently in axes coordinates, so use transAxes to
    # put into data coordinates. *zip(a, b) produces a list of x-coords,
    # then a list of y-coords.
    ax.plot(*zip(location, end), transform=ax.transAxes, **plot_kwargs)

    # Push text away from bar in the perpendicular direction.
    midpoint = (location + end) / 2
    offset = text_offset * np.array([-np.sin(angle_rad), np.cos(angle_rad)])
    text_location = midpoint + offset

    # 'rotation' keyword argument is in text_kwargs.
    ax.text(*text_location, f"{length} {unit_name}", rotation_mode='anchor',
            transform=ax.transAxes, **text_kwargs)



# -*- coding: utf-8 -*-
"""
Created on Sun Dec 13 20:48:05 2015

@author: cdholmes
"""

import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import cartopy.feature as cfeature 
import numpy as np

#########################################################
# Create artificial gridded data

# Number of points in x and y directions
n = 20

# Edges of the latitude-longitude grid
latedge  = np.linspace(-90,90,n+1)
lonedge  = np.linspace(-180,180,2*n+1)
latedge[0] = -89.9999 # Offset south pole to avoid singularity in LambertConformal projection
lon2de, lat2de = np.meshgrid( lonedge, latedge )

# Centers of the latitude-longitude grid
lat      = np.convolve(latedge, [0.5,0.5], 'valid' )
lon      = np.convolve(lonedge, [0.5,0.5], 'valid' )
lon2d, lat2d = np.meshgrid( lon, lat )

# Artificial data
data = np.exp( -lat2d**2/90**2 -lon2d**2/120**2 )

#########################################################
# Create artificial point data

# number of points
npoint = 50

# random latitude longitude and value
plat  = np.random.rand(npoint) * 25 + 25
plon  = np.random.rand(npoint) * 50 - 120
pdata = np.random.rand(npoint) 

##########################################################
# Set up map and plot data

# Open a figure
plt.figure(1)
plt.clf()

# data coordinates: Use Geodetic (or PlateCarree) for data in lat-lon coordinates
datacoord = ccrs.Geodetic()
datacoord = ccrs.PlateCarree()

# Map projection for Global 
ax = plt.axes(projection=ccrs.PlateCarree())
ax = plt.axes(projection=ccrs.Robinson())
ax = plt.axes(projection=ccrs.InterruptedGoodeHomolosine())

# Map projection for North Pole
ax = plt.axes(projection=ccrs.NorthPolarStereo())
ax.set_extent([-180, 180, 60, 90], ccrs.PlateCarree())

# Map projection for CONUS 
ax = plt.axes(projection=ccrs.LambertConformal())
#ax = plt.axes(projection=ccrs.LambertConformal(central_longitude=-96,central_latitude=39,standard_parallels=(33,45)))
# Set lat-lon of map edges. These are reasonable for CONUS
ax.set_extent([-120, -73, 25, 50], ccrs.PlateCarree())

# Add your gridded data
plt.pcolormesh(lonedge, latedge, data, transform=datacoord)

# Add your point data
plt.scatter(plon, plat, marker='o', c=pdata, s=50, transform=datacoord)

# Add a colorbar 
plt.colorbar()

# Add coasts and states; Use ‘110m’ or ‘50m’ resolution
ax.coastlines(resolution='110m')
ax.add_feature(cfeature.STATES.with_scale('110m'))

plt.show()



#!/usr/local/bin/env python3
''' Module to calculate solar zenith angle, solar declination, and equation of time 
Results should be accurate to < 0.1 degree, but other modules should be used for 
high-precision calculations.

C.D. Holmes 9 Nov 2018
'''

import numpy as np
import pandas as pd

pi180 = np.pi / 180

def solar_declination( date ):
    '''Calculate solar declination (degrees) for specified date
    
    Implements Eq. 9.68-9.72 from M.Z. Jacobson, Fundamentals of Atmospheric Modeling
    
    Argument
    --------
    date : pandas.Timestamp, date, or datetime
        date for calculation

    Returns
    -------
    dec : float
        solar declination in degrees at the specified date
    
    # C.D. Holmes - 9 Nov 2018 - Initial version
    '''

    # Convert to pandas Timestamp, if needed
    if not isinstance(date, pd.Timestamp):
        date = pd.Timestamp(date)

     # Number of days since beginning of 2000
    NJD = np.floor( date.to_julian_date() - pd.Timestamp(2000,1,1).to_julian_date() )

    # Obliquity, degrees
    ob = 23.439 - 4e-7 * NJD

    # Parameters for ecliptic, degrees
    gm = 357.528 + 0.9856003 * NJD
    lm = 280.460 + 0.9856474 * NJD

    # Ecliptic longitude of sun, degrees
    ec = lm + 1.915 * np.sin( gm * pi180 ) + 0.020 * np.sin( 2 * gm * pi180 )

    #Solar declination, degrees
    dec = np.arcsin( np.sin( ob * pi180 ) * np.sin( ec * pi180 ) ) / pi180

    return dec

def equation_of_time( date ):
    '''Calculate equation of time (degrees) for specified date
    
    Implements the "alternative equation" from Wikipedia, derived from
    https://web.archive.org/web/20120323231813/http://www.green-life-innovators.org/tiki-index.php?page=The%2BLatitude%2Band%2BLongitude%2Bof%2Bthe%2BSun%2Bby%2BDavid%2BWilliams
    Results checked against NOAA solar calculator and agree within 10 seconds.
    
    Argument
    --------
    date : pandas.Timestamp, date, or datetime
        date for calculation

    Returns
    -------
    eot : float
        equation of time in degrees on the specified date

    # C.D. Holmes - 9 Nov 2018 - Initial version
    '''
    # Convert to pandas Timestamp, if needed
    if not isinstance(date, pd.Timestamp):
        date = pd.Timestamp(date)

    # Equation of time, accounts for the solar day differing slightly from 24 hr
    doy = date.dayofyear
    W = 360 / 365.24
    A = W * (doy+10)
    B = A + 1.914 * np.sin( W * (doy-2) * pi180 )
    C = ( A - np.arctan2( np.tan(B*pi180), np.cos(23.44*pi180) ) / pi180 ) / 180

    # Equation of time in minutes of an hour
    eotmin = 720 * ( C - np.round(C) )

    # Equation of time, minutes -> degrees
    eot = eotmin / 60 * 360 / 24

    return eot

def hour_angle( lon, datetimeUTC ):
    '''Compute hour angle (degrees) for specified longitude, date and time

    Hour angle is the angular displacement of the sun from the local meridian.
    It is zero at local noon, negative in the morning, and positive is afternoon.
    
    Parameters
    ----------
    lon : float
        longitude in degrees east
    datetimeUTC : pandas.Timestamp or datetime
        date and time for calculation, must be UTC
    
    Returns
    -------
    ha : float
        hour angle in degrees at the specified location and time
    '''

    # Convert to pandas Timestamp, if needed
    if not isinstance(datetimeUTC, pd.Timestamp):
        datetimeUTC = pd.Timestamp(datetimeUTC)

    # Hour angle for mean solar time.
    # Actual solar position has a small offset given by the equation of time (below)
    Ha = ( datetimeUTC.hour + datetimeUTC.minute / 60 - 12 ) * 15 + lon

    # Add equation of time to the hour angle, degrees
    Ha += equation_of_time( datetimeUTC )

    return Ha

def refraction_angle( true_elevation_angle, pressure=101325., temperature_celsius=10. ):
    '''Atmospheric refraction angle

    apparent elevation angle = (true elevation angle) + (refraction angle)
    Equation from Saemundsson/Bennett

    Parameters
    ----------
    true_elevation_angle : float
        degrees above horizon of sun or other object
    pressure : float
        surface atmospheric pressure (Pa)
    temperature_celsius : float
        surface atmospheric temperature (C)

    Returns
    -------
    angle : float
        refraction angle in degrees. Value is zero when apparent elevation is below horizon
    '''
    # Refraction angle, arcminutes
    R = 1.02 / np.tan( ( true_elevation_angle + 10.3 / (true_elevation_angle + 5.11) ) * pi180 )
    # Account for temperature and pressure, arcminutes
    R = R * pressure / 101325 * 283 / ( 273 + temperature_celsius )
    # Convert arcminutes -> degrees
    R /= 60

    # Result must be positive
    R = np.maximum(R,0)

    # Refraction defined only when the apparent elevation angle is positive
    # Set refraction to zero when the apparent elevation is below horizon
    refraction_angle = np.where( true_elevation_angle + R <= 0, 0, R)

    return refraction_angle

def solar_zenith_angle( lat, lon, datetimeUTC, 
                       refraction=False, temperature=10., pressure=101325. ):
    '''Calculate solar zenith angle for a given latitude, longitude, date and time.
    
    Accounts for equation of time and (optionally) for atmospheric refraction.
    Altitude of the observer is not accounted for, which can be important when the sun 
    is near the horizon. 
    
    Results are accurate to tenths of a degree, except where altitude is important
    (< 20 degrees solar elevation)

    Parameters
    ----------
    lat : float or ndarray
        latitude in degrees
    lon : float or ndarray
        longitudes in degrees
    datetimeUTC : pandas.Timestamp or datetime
        date and time in UTC
    refraction : bool (default=False)
        specifies whether to account for atmospheric refraction
    temperature : float or ndarray
        surface atmospheric temperature (Celsius), only used for refraction calculation
    pressure : float or ndarray
        surface atmospheric pressure (Pa), only used for refraction calculation
    
    Returns
    -------
    sza : float or ndarray
        solar zenith angle in degrees at the designated locations and times
        If refraction=False, this is the true solar zenith angle
        If refraction=True, this is the apparent solar zenith angle

    C.D. Holmes - 9 Nov 2018 - Initial version
    '''
    # Convert to pandas Timestamp, if needed
    if not isinstance(date, pd.Timestamp):
        date = pd.Timestamp(date)

    # Solar declination, degrees
    dec = solar_declination( datetimeUTC )

    # Hour angle, degrees
    Ha = hour_angle( lon, datetimeUTC )

    # True solar zenith angle, radians
    sza = np.arccos( np.sin(lat*pi180) * np.sin(dec*pi180) + \
          np.cos(lat*pi180) * np.cos(dec*pi180) * np.cos(Ha*pi180) )

    # Convert radians -> degrees
    sza /= pi180

    if refraction:
        # Subtract refraction angle (degrees) from zenith angle.
        # SZA is always smaller due to refraction.
        sza -= refraction_angle( 90-sza, pressure, temperature )

    return sza

def solar_azimuth_angle( lat, lon, datetimeUTC ):
    '''Solar azimuth angle (degrees) for a latitude, longitude, date and time
    
    SAA is degrees clockwise from north.
    
    Parameters
    ----------
    lat : float or ndarray
        latitude in degrees
    lon : float or ndarray
        longitudes in degrees
    datetimeUTC : pandas.Timestamp or datetime object
        date and time in UTC

    Returns
    -------
    saa : float or ndarray
        solar azimuth angle in degrees (clockwise from north)

    C.D. Holmes - 13 Jan 2023 - Initial version
    '''
    # Convert to pandas Timestamp, if needed
    if not isinstance(date, pd.Timestamp):
        date = pd.Timestamp(date)

    # Solar declination, degrees
    dec = solar_declination( datetimeUTC )

    # Hour angle, degrees
    Ha = hour_angle( lon, datetimeUTC )

    # Solar zenith angle, degrees
    # Use true sza, without refraction
    zen = sza( lat, lon, datetimeUTC, refraction=False )

    # Solar azimuth angle, degrees
    saa = np.arcsin( -np.sin( Ha*pi180 ) * np.cos( dec*pi180 ) / 
            np.sin( zen*pi180 ) ) / pi180

    # Change range [-180,180] to [0,360]
    return np.mod( saa+360, 360 )

# Aliases for functions
sza = solar_zenith_angle
saa = solar_azimuth_angle
# Additional aliases for backwards compatibility
equationOfTime = equation_of_time
solarDeclination = solar_declination


#!/usr/bin/env python3
'''Demonstrate conversion between geodetic coordinates (latitude, longitude) 
and map projection coordinates (x,y)'''

#%%
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import cartopy.feature as cfeature 
import numpy  as np
import pandas as pd

# Define a few points in latitude, longitude
points_usa    = pd.DataFrame( {'lat':[40,45],'lon':[-70,-80]} )
points_arctic = pd.DataFrame( {'lat':[80,85],'lon':[10,-90]} )

# Define coordinate reference systems
# Note that for some applications, shape of the Earth should be specified 
geo = ccrs.Geodetic()
nps = ccrs.NorthPolarStereo()

# Convert geodetic coordinates to projection coordinates
points = nps.transform_points(geo,
                            points_arctic['lon'].values,
                            points_arctic['lat'].values )
# Add results to data frame
points_arctic['x'] = points[:,0]
points_arctic['y'] = points[:,1]

#%%
# Plot the points on a map
plt.figure(1)
plt.clf()

# Map projection for North Pole
ax = plt.axes(projection=ccrs.NorthPolarStereo())
ax.set_extent([-180, 180, 60, 90], ccrs.PlateCarree())

# Add coasts and states; Use ‘110m’ or ‘50m’ resolution
ax.coastlines(resolution='110m')
ax.add_feature(cfeature.STATES.with_scale('110m'))

# Plot points with geodetic coordinates
plt.scatter(points_arctic['lon'], 
            points_arctic['lat'], 
            marker='v', s=50, 
            transform=ccrs.PlateCarree())

# Plot points with projection coordinates, showing results are same
plt.scatter(points_arctic['x'], 
            points_arctic['y'], 
            marker='^', s=50, 
            transform=nps)

plt.show()
# %%
